from .postgres_handler import Postgres_Handler as pg_handler

class Database_Handler:
    def __init__(self, db_type, user_name, password, host, database, port):
        self.db_type = db_type
        self.__build_connection(user_name, password, host,database, port)

    def __build_connection(self, user_name, password, host, database, port):
        self.handler = pg_handler(user_name, password, host, database, port)
    
    def search_records(self, columns, keywords):
        return self.handler.search_for_records(columns, keywords)
    
    def insert_record(self, **keywords):
        return self.handler.insert_row(**keywords)
    
    def update_record(self, keywords):
        return self.handler.update_record(keywords)
    


from psycopg2 import sql
import psycopg2
import json

class Postgres_Handler:
    #Key for table_name string
    table_name_key = 'table_name'
    
    #Initializer that takes the values for connection
    def __init__(self, user_name, password, host, database, port):
        self.user_name = user_name
        self.password = password
        self.host = host
        self.port = port
        self.database = database
        
   
    # Used internally create to run query
    def __run_query(self, query, values):
        #Sets the message to 'No records found' and status -1
        status = 0
        results = 'No records found'
        
        #Creates cursor to iterate through rows
        with psycopg2.connect(user=self.user_name, password=self.password, host=self.host, dbname=self.database) as conn:
            with conn.cursor() as cursor:
        
                try:
                    print(query)
                    cursor.execute(query, values)
                    results = cursor.fetchall()

                    status = 1
                    conn.commit()
                    
                    if results.count == 0:
                        results = 'No records found'
                        
                except Exception as ex:
                    status = -1
                    results = f"{ex}"
            
        return (status, results)
    
    #Method to close the connection    
    def Close(self):
            try:
                self.conn.close() 
            except Exception as ex:
                return f'Error: {ex}'
        
    #Select all rows 
    def search_for_records(self, columns, keywords):
        payload = {}
    
        query_value_list = self.__build_search_query(keywords)       

        #Runs the query and returns the results
        results = self.__run_query(query_value_list[0], query_value_list[1])
        payload = self.__format_search_results(results, columns)
    
        return payload

    def __build_search_query(self, keywords):
        where_string = ''
        values_dict = {}
        table_name_key = 'table_name'
        
        #Start of query string
        query = f'SELECT * FROM "{keywords[table_name_key]}"'
        
        if table_name_key not in keywords:
            assert("Error: table_name required as argument")
        values_dict[table_name_key] = keywords[table_name_key]
        
        #Keys for the StartDate and EndDate
        start_date_key = "SearchStartDate"
        end_date_key = "SearchEndDate"

        if len(keywords) != 0:
            #Iterates through the keywords to build search string
            for key, item in keywords.items():
                if key == table_name_key: continue
                
                if where_string != '':
                    where_string += ' AND '
                
                if key == start_date_key:
                    where_string += f'"DocumentDate" >= %({key})s'
                    values_dict[key] = item
                elif key == end_date_key:
                    where_string += f'"DocumentDate" <= %({key})s'
                    values_dict[key] = item
                else:
                    where_string += f'"{key}" ILIKE %({key})s'
                    values_dict[key] = f'{item}'
                    
            if where_string != '':
                query = f"{query} WHERE {where_string}"
                
        return [query, values_dict]
        
    
    def __format_search_results(self, results, columns):
        formatted_rows = []
        records = ''
        error = ''
        
        status_code = 200
        internal_status = results[0]
        data = results[1]  
        
        if internal_status != -1:
            row_count = len(data)
            
            if row_count > 0:
                for row in data:
                    formatted_row = {}
                    
                    for index, item in enumerate(row):
                        formatted_row[columns[f"{index}"]] = f'{item}'
                    
                    formatted_rows.append(formatted_row)

            records = {
                'count': row_count,
                'records': formatted_rows
            }
        else:
            error = data
            
        return {
            'statusCode': status_code,
            'headers': {
                'Content-Type': "application/json",
                'Access-Control-Allow-Headers': 'Content-Type',
                'Access-Control-Allow-Origin': '*',
                'Access-Control-Allow-Methods': 'OPTIONS,POST,GET',
            },
            "body": json.dumps({
                'results': records,
                'error' : f'{error}'
            })
        } 
        
        
    #This needs to be looked at, likely there are only certain fields that will be passed and the rest will be created either in lambda or S3/postgresql
    def insert_row(self, **kwargs):
        
        #Values dictioniary for sql parameters
        # values_dict = {}
        
        # #Empty strings for column names and column values to be added
        # column_names_string = ''
        # column_values_string = ''
        
        #checks for the table name in args
        if self.table_name_key not in kwargs:
            assert("Error: table_name required as argument")
       
        query_dict = {key: val for key,val in kwargs.items() if key != self.table_name_key}

        #Begining of the SQL query
        query = sql.SQL('INSERT INTO {table_name} ({columns}) VALUES({values})').format(
                columns = sql.SQL(', ').join(
                    map(sql.Identifier, query_dict.keys())
                ),
                values = sql.SQL(', ').join(
                    map(sql.Placeholder, query_dict.keys())
                ),
                table_name = sql.Identifier(kwargs[self.table_name_key])
            )
        
    
        
        #Iterates through the arguments adds the key to the column_name, and the value to the column_value
        # for key, value in kwargs.items():
        #     column_names_string += f'"{key}",'
        #     column_values_string += f"%({key})s,"
        #     values_dict[key] = value
            
        # column_names = column_names_string[:-1]
        # column_values = column_values_string[:-1]
        
        # #Builds the final SQL query for updating values
        # query = f'{query_start}({column_names}) VALUES({column_values})'
        # query = query.format(table_name = sql.Identifier(table_name))
        #Runs the command
        results = self.__run_query(query, kwargs)
        return self.__format_results(results)
        
    def __format_results(self, results):
        internal_status = results[0]
        if internal_status == -1:
            print(results[1])
            error = results[1]
            message = ''
            status_code = 400
        else:
            error = ''
            message = results[1]
            status_code = 200
                
        return {
            'statusCode': status_code,
            'headers': {
                'Content-Type': "application/json",
                'Access-Control-Allow-Headers': 'Content-Type',
                'Access-Control-Allow-Origin': '*',
                'Access-Control-Allow-Methods': 'OPTIONS,POST,GET',
            },
            "body": json.dumps({
                'results': f'{message}',
                'error' : f'{error}'
            })
        } 
        
    #Method to update a SQL row
    def update_record(self, kwargs):
        values_dict = {}
        id_key = 'id'
        clause_section = ''
        
         #checks for the table name in args
        if self.table_name_key not in kwargs:
            assert("Error: table_name required as argument")
        table_name = kwargs[self.table_name_key]
        
        #Begining of the SQL query
        query_start = 'UPDATE %(table_name)s SET'
        values_dict[self.table_name_key] = table_name
        
        for key, value in kwargs.items():
            if key.lower() == "id": continue
            
            clause_section = f' "{key}" = %({key})s,'
            values_dict[key] = value
        
        query_string += clause_section[:-1]
        
        if id_key not in kwargs: assert("Error: ID of record required for an update!")
        id_val = kwargs['id']
        query_string += f' WHERE "ID"=\'{id_val}\''
        
        results = self.__run_query(query_string)
        return self.__format_results(results)
    
        
    
    #Delete command, will use policy or agent number to delete
    #todo - This may requrie removing items from S3 not sure yet how that will work
    def deleteRow(self, **kwargs):
        #key for table and id 
        table_name_key = "table_name"
        id_key = "ID"
        
        value_dict = {key: value for key, value in kwargs.items() if key.lower() == table_name_key.lower() or key.lower() == id_key.lower()}
        
        #Builds query to run 
        query_string = 'DELETE FROM %(table_name)s WHERE ID = %(ID)s'
        
        #Runs the query and handles the success or error.
        results = self.__run_query(query_string, value_dict)
        return results



from abc import ABC, abstractmethod

class Sql_Handler(ABC):
    @abstractmethod
    def define_columns(self, columns):
        self.columns = columns
    
    @abstractmethod
    def set_table_name(self, table_name):
        self.table_name = table_name
        
    @abstractmethod
    def set_keywords(self, keywords):
        self.keywords = keywords


from .connectiontypes.database_handler import Database_Handler

def get_handler(db_type, user_name, password, host, database, port):
    return Database_Handler(db_type, user_name, password, host, database, port)


import boto3
from ..models.scan_object import Scan_Object

class Dynamo_Handler:
    def __init__(self):
        self.client = boto3.client("dynamodb")
    
    def add_item(self, **kwargs):
        table_name_key = "table_name"
        fields_key = "fields"
        
        if table_name_key not in kwargs:
            assert("Error: Table name must be provided to create a new entry")    
        table_name = kwargs[table_name_key]
        
        if fields_key not in kwargs:
            assert("Error: Must provide fields for new entry")
        fields = kwargs[fields_key]
        
        name_key = 'item_name'
        type_key = 'item_type'
        val_key = 'item_value'
         
        item = {f'{field[name_key]}': {f'{field[type_key]}': f'{field[val_key]}'} for field in fields }
        
        self.client.put_item(
            TableName = table_name,
            Item=item
        )
    
    def get_item(self, table_name, partition, sort_field = ""):
        keyword_name_key = 'name'
        keyword_value_key = 'value'
        keyword_type_key = 'type'
        
        if keyword_name_key not in partition:
            assert(f"Error: {keyword_name_key} missing in partition argument")
        if keyword_type_key not in partition:
            assert(f"Error: {keyword_type_key} missing in partition argument")
        if keyword_value_key not in partition:
            assert(f"Error: {keyword_value_key} missing in partition argument")
        
        key = {partition[keyword_name_key]: {partition[keyword_type_key]: partition[keyword_value_key]}}
        if sort_field != '':
            
            if keyword_name_key not in sort_field:
                assert(f"Error: {keyword_name_key} missing in sort_field argument")
            if keyword_type_key not in sort_field:
                assert(f"Error: {keyword_type_key} missing in sort_field argument")
            if keyword_value_key not in sort_field:
                assert(f"Error: {keyword_value_key} missing in sort_field argument")
            
            
            key[sort_field[keyword_name_key]] = {sort_field[keyword_type_key]: sort_field[keyword_value_key]}
            
        response = self.client.get_item(
            TableName = table_name,
            Key = key
        )
        
        return response
         
    def create_table(self, **kwargs):
        #List of keyword keys required for create a table
        table_key = "table_name"
        schema_key = "key_schema"
        attributes_key ="attributes"
        provisioned_key = "provisioned"
        
        #Checks to make sure required keys are passed to create_table method
        keys = [table_key, schema_key, attributes_key, provisioned_key]
        for key in keys:
            if key not in kwargs:
                assert(f"Error: Please provide {key} as an argument")
    
        #Retrieves the arguments from kwargs
        table_name = kwargs[table_key]
        schema = kwargs[schema_key]
        attributes = kwargs[attributes_key]
        provisoned = kwargs[provisioned_key]
        
        #Creates the schema and attribute dictionary 
        schema_list = [{'AttributeName': key, 'KeyType': val} for key, val in schema.items()]
        attribute_list = [{'AttributeName': key, 'AttributeType': val} for key, val in attributes.items()]
        
        #creates the new table with the elemets provided
        table = self.client(
            TableName = table_name,
            KeySchema=schema_list,
            AttributeDefinitions=attribute_list,
            ProvisionedThroughput=provisoned
        )                
        
        #waits and confirms table exists
        table.wait_until_exists()
        
        return table.item_count
    
    def scan_table(self, table_name, scan_filters, display_filters=""):
        scan = self.__build_scan(table_name, scan_filters, display_filters)
        items = scan.get_scan_object()
        
        return self.client.scan(**items)
        
    def __build_scan(self, table_name, scan_filters, display_filters=""):
        request_dict = {}
        
        attribute_short_key = "short_name"
        attribute_name_key = "name"
        attribute_value_key = "value"
        
        if len(scan_filters) == 0:
            assert("Error: scan_filters can not be empty")
        
        #Checks for keys and assigns the value
        if attribute_name_key not in scan_filters:
            assert(f"Error: {attribute_name_key} required for scan_filters")
        scan_filter_name = scan_filters[attribute_name_key]
        
        if attribute_short_key not in scan_filters:
            assert(f"Error: {attribute_short_key} required for scan_filters")
        scan_filter_short_name = scan_filters[attribute_short_key]
        
        if attribute_value_key not in scan_filters:
            assert(f"Error: {attribute_value_key} required for scan_filters")
        scan_filter_value = scan_filters[attribute_value_key]
        
        #creats the sca_filter_dict for the ExpressionAttributeValues
        scan_filter_dict = {scan_filter_short_name: scan_filter_value}
        #creates the scan_filter_string for the FilterExpression
        scan_filter_string = f'{scan_filter_name} = {scan_filter_short_name}'
        
        #Creats the scan_object and adds the initial values
        scan_object = Scan_Object(table_name, scan_filter_dict, scan_filter_string)
        
        #Checks to see if display filters exist
        if display_filters != "":
            attribute_names = {}
            expression_string = ""
            
            #Loops through filters to confirm there are not missing keys
            for filter in display_filters:
                if attribute_short_key not in filter:
                    assert(f"Errror: [{attribute_name_key}] required for all display_filters.")
                if attribute_name_key not in filter:
                    assert(f"Errror: [{attribute_short_key}] required for all display_filters.")
                        
                attribute_names[attribute_short_key] = attribute_name_key
                expression_string += f'{filter[attribute_short_key]},'
            
            expression_string = expression_string[:-1]
            
            #Creates the expression attribute names and ProjectExpression string
            scan_object.ExpressionAttributeNames = attribute_names
            scan_object.ProjectionExpression = expression_string
        
        #returns the scan_object to initiate the scan
        return scan_object
        
        
        
        